/*
 * ar3cmonitor - analogcache.cpp
 *
 * 	Analog Cache
 *
 * Copyright (C) 2007..2011  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "ar3cmonitor.h"
#include "analogcache.h"
#include "control.h"

AnalogCache::AnalogCache(unsigned int phys)
	: quantum(0), vquantum(0), physical(phys), need_update(1),
	want_optimize(0), ttl(ANALOG_CACHE_TTL), count(0UL), add_counter(0),
	touched(1)
{
	INIT_LIST_HEAD(&head.list);
	head.is_head = 1;
	head.ts = 0;
	ACC->register_cache(physical, this);
}

void AnalogCache::update_all()
{
	unsigned long long end = CO->cursorStamp();
	unsigned long long begin = end - CO->timeRange();
	unsigned int i = 0;
	int v = 0;

	AnalogControl *AC = CO->analog();
	AList::AIterator it(AC->Cursor, physical);
	it.positionRight();
	if (!it.valid())
		--it;

	while (likely(it.valid())) {
		unsigned long long t = it.timeStamp();
		struct ac_elem *e = (struct ac_elem *) malloc(sizeof(*e));
		int ret = 0;

		e->ts = t;
		e->is_head = 0;

		if (unlikely(analog_get(it, physical, &v))) {
			e->defined = 0;
			v = 999999;
		} else {
			e->defined = 1;
			e->value = v;
		}

		list_add(&e->list, &head.list);
		count++;
		i++;

		if (t < begin)
			break;

		for (;;) {
			int iv;

			--it;
			if (unlikely(!it.valid()))
				break;

			if (unlikely(it.timeStamp() <= t - quantum))
				break;

			ret = analog_get(it, physical, &iv);
			if (ret)
				continue;

			if (abs(iv - v) >= vquantum)
				break;
		}
	}

	optimize();
	touch();
	need_update = 0;
}

void AnalogCache::update_cache(unsigned long long ts, short value)
{
	unsigned long long cursor = CO->cursorStamp();
	unsigned long long left = cursor - CO->timeRange();
	struct ac_elem *e;
	unsigned long long end = end_ts();

	if (need_update)
		return;

	if (!ttl) {
#if DEBUG
		qWarning("cache died: %d", physical);
#endif
		clear();
		need_update = 1;
	}

	ttl--;

	e = get_last();
	if (e->ts <= cursor) {
		if ((ts > left) && (ts < e->ts)) {
			need_update = 1;
			return;
		}
		if ((ts < end + quantum) && (abs(e->value - value) < vquantum))
			return;

		struct ac_elem *e = (struct ac_elem *) malloc(sizeof(*e));

		e->ts = ts;
		e->is_head = 0;
		e->defined = 1;
		e->value = value;

		list_add_tail(&e->list, &head.list);
		count++;

		/* optimize() on all caches might take some time */
		if (add_counter++ > (256 + ((((unsigned long) this) >> 4)
						& 0xff))) {
			want_optimize = 1;
			add_counter = 0;
		}
		touch();
		return;
	}
	if ((ts > left) && (ts < cursor))
		need_update = 1;
}

#ifdef CACHE_STATS
#ifdef HAVE_SYS_TIME_H

#include <sys/time.h>

static unsigned long long time_us(void)
{
	struct timeval tv;

	gettimeofday(&tv, NULL);
	return tv.tv_sec * 1000000ULL + tv.tv_usec;
}

#else

static unsigned long long time_us(void)
{
	return clock() * 1000000ULL / CLOCKS_PER_SEC;
}

#endif
#endif

extern int acache_opt;

int AnalogCache::ge(struct ac_elem *a, struct ac_elem *b)
{
	return (int) a->value >= (int) b->value - vquantum;
}

void AnalogCache::optimize()
{
	struct ac_elem *a, *b, *c, *d, *e;
#ifdef CACHE_STATS
	unsigned long prev_count = count;
#endif

	want_optimize = 0;

	/* Don't optimize small caches */
	if (!acache_opt || count < 512)
		return;

#ifdef CACHE_STATS
	unsigned long long t = time_us();
#endif
	e = get_last();
	d = list_entry(list_prev(&e->list), struct ac_elem, list);
	c = list_entry(list_prev(&d->list), struct ac_elem, list);
	b = list_entry(list_prev(&c->list), struct ac_elem, list);
	a = list_entry(list_prev(&b->list), struct ac_elem, list);

	for (; !list_is_head(&a->list, &head.list);
			d = list_entry(list_prev(&e->list), struct ac_elem,
				list),
			c = list_entry(list_prev(&d->list), struct ac_elem,
				list),
			b = list_entry(list_prev(&c->list), struct ac_elem,
				list),
			a = list_entry(list_prev(&b->list), struct ac_elem,
				list)) {

		/* 3-point optimization */
		if (!(a->defined && b->defined && c->defined)) {
			e = d;
			continue;
		}

		if ((a->value == b->value) && (b->value == c->value)) {
			list_del(&b->list);
			free(b);
			count--;
			continue;
		}

		if (c->ts - a->ts > quantum) {
			e = d;
			continue;
		}

		if ((ge(b, a) && ge(c, b)) || (ge(a, b) && ge(b, c))) {
			list_del(&b->list);
			free(b);
			count--;
			continue;
		}

		/* 4-point optimization */
		if (!d->defined || (d->ts - a->ts > quantum)) {
			e = d;
			continue;
		}

		if ((ge(b, a) && ge(b, c) && ge(d, b))
				|| (ge(a, b) && ge(c, b) && ge(b, d))) {
			list_del(&b->list);
			free(b);
			count--;
			continue;
		}

		e = d;
	}

#ifdef CACHE_STATS
	t = time_us() - t;

	qWarning("AnalogCache opt.: %ld->%ld (%ld%%) in %Ld.%03Ldms",
			prev_count, count, (count * 100) / prev_count,
			t / 1000, t % 1000);
#endif
}

void AnalogCache::update()
{
	if (need_update) {
		clear();
		update_all();
	} else if (want_optimize)
		optimize();
	ttl = ANALOG_CACHE_TTL;
}

void AnalogCache::set_quantum(unsigned long long q)
{
	if (quantum > q) {
		quantum = q;
		need_update = 1;
	} else if (quantum < q) {
		quantum = q;
		want_optimize = 1;
	}
}

void AnalogCache::set_vquantum(int vq)
{
	if (vquantum > vq) {
		vquantum = vq;
		need_update = 1;
	} else if (vquantum < vq) {
		vquantum = vq;
		want_optimize = 1;
	}
}

void AnalogCache::invalidate_range(unsigned long long from,
		unsigned long long to)
{
	unsigned long long right = CO->cursorStamp();
	unsigned long long left = right - CO->timeRange();

	if (need_update)
		return;

	if ((to >= left) && (from <= right)) {
		clear();
		need_update = 1;
	}
}

void AnalogCache::invalidate(void)
{
	if (!need_update) {
		clear();
		need_update = 1;
	}
}

void AnalogCache::cleanup(ac_elem *e)
{
	ac_elem *tmp;

	e = list_entry(list_prev(&e->list), ac_elem, list);

	while (!list_is_head(&e->list, &head.list)) {
		tmp = e;
		e = list_entry(list_prev(&e->list), ac_elem, list);

		list_del(&tmp->list);
		count--;
		free(tmp);
	}
	touch();
}

void AnalogCache::clear()
{
	ac_elem *e;
	while (!list_empty(&head.list)) {
		e = list_entry(list_next(&head.list), ac_elem, list);

		list_del(&e->list);
		count--;
		free(e);
	}

	BUG_ON(count);
	touch();
}

AnalogCache::~AnalogCache()
{
	ACC->unregister_cache(this);
}

AnalogCacheControl *ACC;

AnalogCacheControl::AnalogCacheControl(unsigned int n)
	: count(n)
{
	unsigned int i;

	caches = new list_head[count];
	for (i = 0; i < count; i++)
		INIT_LIST_HEAD(caches + i);

#ifdef CACHE_STATS
	caches_count = 0;
#endif
}

void AnalogCacheControl::register_cache(unsigned int phys, AnalogCache *c)
{
	BUG_ON(phys >= count);
	c->cache_list.cache = c;
	list_add(&c->cache_list.list, caches + phys);
#ifdef CACHE_STATS
	caches_count++;
#endif
}

void AnalogCacheControl::unregister_cache(AnalogCache *c)
{
	list_del(&c->cache_list.list);
#ifdef CACHE_STATS
	caches_count--;
#endif
}

void AnalogCacheControl::stats()
{
#ifdef CACHE_STATS
	qWarning("analogcache stats: count: %d", caches_count);
#endif
}

void AnalogCacheControl::invalidate_all()
{
	unsigned int i;
	struct acache_list_head *c;

	for (i = 0; i < count; i++)
		list_for_each_entry(c, caches + i, list)
			c->cache->invalidate();
}

void AnalogCacheControl::invalidate_range_all(unsigned long long from,
		unsigned long long to)
{
	unsigned int i;
	struct acache_list_head *c;

	for (i = 0; i < count; i++)
		list_for_each_entry(c, caches + i, list)
			c->cache->invalidate_range(from, to);
}

void AnalogCacheControl::update_cache(unsigned int phys,
		unsigned long long ts, short value)
{
	struct acache_list_head *c;

	BUG_ON(phys >= count);
	list_for_each_entry(c, caches + phys, list)
		c->cache->update_cache(ts, value);
}


void AnalogCacheControl::invalidate_range(unsigned int phys,
		unsigned long long from, unsigned long long to)
{
	struct acache_list_head *c;

	BUG_ON(phys >= count);
	list_for_each_entry(c, caches + phys, list)
		c->cache->invalidate_range(from, to);
}

void AnalogCacheControl::invalidate(unsigned int phys)
{
	struct acache_list_head *c;

	BUG_ON(phys >= count);
	list_for_each_entry(c, caches + phys, list)
		c->cache->invalidate();
}
