//
// C++ Implementation: control
//
// Description:
//
//
// Author: Szymon Acedanski <accek@podlesie.net>, (C) 2004
// Copyright (C) 2005, 2006, 2007  Krzysztof Mazur <krzysiek@podlesie.net>
//
// Copyright: See COPYING file that comes with this distribution
//
//

#include <cstdio>
#include <algorithm>
#include <qglobal.h>
#include <qtimer.h>
#include <errno.h>
#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif

#include "ar3cmonitor.h"
#include "control.h"
#include "commobj.h"
#include "consts.h"
#include "analogcache.h"
#include <fonts.h>

ControlObject *CO = NULL;

extern unsigned long mem_low, mem_high;
unsigned long long off_reqs, off_time;
extern unsigned int ar3cmon_update_interval;
extern unsigned int ar3cmon_offline_update_interval;

static QFont make_font(enum ar3cmonitor_font font)
{
	const char *family;
	unsigned int size;
	QFont f;

	family = ar3cmon_font_family(font);
	size = ar3cmon_font_size(font);

	if (family != NULL) {
		f.setStyleHint(QFont::SansSerif, QFont::PreferBitmap);
		f.setFamily(family);
	}
	f.setPointSize(size);
	return f;
}

ControlObject::ControlObject(int nbinaries)
	: nBinaries(nbinaries),
	Online(nBinaries, "Online Storage"),
	Offline(nBinaries, "Offline Storage"),
	TimeStamp(0ULL),
	cursor_mode(CURSOR_MODE_LRM)
{
	Analog = Online.get_acontrol();
	Binary = Online.get_bcontrol();
	BObjects = new BinaryObject[nBinaries];
	active = &Online;
	TimeRange = AR3CMON_DATA_TIME_RANGE;
	OfflineMode = false;
	OfflineDownloading = false;
	HaveData = false;
	ReviewMode = false;
	SetupHash = 0;
	fm_req = lm_req = NULL;
	VGridColor = QColor(0, 0, 90, QColor::Hsv);
	HGridColor = VGridColor;

	HugeFont = make_font(AR3CMONITOR_FONT_HUGE);
	ValueFont = make_font(AR3CMONITOR_FONT_VALUE);
	SmallFont = make_font(AR3CMONITOR_FONT_SMALL);
	TimescaleFont = make_font(AR3CMONITOR_FONT_TIMESCALE);

	maxBinUsed = -1;
	gcollector_time = 0;
	gcollector_iter = 0;
	offline_reqs = 0;
	offline_time = 0;
	prev_used = 0;
	lavg = 0;
	layer = -1;

	reset();

	autoAdjustVGrid();

	AdjustCursorsTimer = new QTimer(this);
	connect(AdjustCursorsTimer, SIGNAL(timeout()), SLOT(doAdjustCursors()));

	DataTimer = new QTimer(this);
	connect(DataTimer, SIGNAL(timeout()), SLOT(noData()));

	QTimer* g = new QTimer(this);
	connect(g, SIGNAL(timeout()), SLOT(garbageCollector_stats()));
	g->start(AR3CMON_GARBAGE_INTERVAL);

	QTimer *t = new QTimer(this);
	connect(t, SIGNAL(timeout()), SLOT(watchdog()));
	t->start(1000);

	UpdateInterval = ar3cmon_update_interval;
	UpdateTimer = new QTimer(this);
	connect(UpdateTimer, SIGNAL(timeout()), SIGNAL(sigUpdateTimer()));
	UpdateTimer->start(UpdateInterval);
}

ControlObject::~ControlObject()
{
	delete []BObjects;
}

BinaryControl *ControlObject::onlineBinary(unsigned int id)
{
	return Online.get_binary(id);
}

BinaryControl *ControlObject::offlineBinary(unsigned int id)
{
	return Offline.get_binary(id);
}

void ControlObject::newTimeStamp(unsigned long long ts)
{
	TimeStamp = ts;

	if (!ReviewMode && (ts > cursorStamp()))
		setCursorStamp(ts);
}

int ControlObject::newARecord(unsigned long long ts, ARecord *r)
{
	bool crossed;

	if (!ts)
		return 1;

	newTimeStamp(ts);
	/* po co to? */
       	crossed = TimeStamp < cursorStamp() && cursorStamp() <= ts;
	Online.newARecord(ts, r);
	emit sigNewTimestamp();

	if (crossed)
		emit sigPleaseUpdate();
	return 0;
}

int ControlObject::newOffARecord(unsigned long long ts, ARecord *r)
{
	Offline.newARecord(ts, r);
	adjustCursors();
	return 0;
}

void ControlObject::beginUndefinedOffline(unsigned long long ts,
		unsigned int flags)
{
	Offline.beginUndefined(ts, flags);
}

void ControlObject::newBRecord(unsigned long long ts, int id, BState state)
{
	bool crossed = false;
	BinaryControl *c = onlineBinary(id);
	if (c == NULL)
		return;

	if (ts) {
		newTimeStamp(ts);

		/* po co to? */
       		crossed = TimeStamp < cursorStamp() && cursorStamp() <= ts;
	} else
		c->clear();

	BState ps = Undefined;
	if (c->InsertIt.valid()) {
		BRecord *prev = c->InsertIt.data();
		ps = prev->state;
	}

	if (ps == state)
		return;

	Online.newBRecord(ts, id, state);
	emit bobject(id)->EmitChange();

	if (crossed)
		emit sigPleaseUpdate();

	garbageCollector();
}

void ControlObject::newOffBRecord(unsigned long long ts, int id, BState state)
{
	if (!ts)
		return;
	Offline.newBRecord(ts, id, state);
	adjustCursors();
}

void ControlObject::adjustCursors()
{
	if (!AdjustCursorsTimer->isActive())
		AdjustCursorsTimer->start(100, TRUE);
}

void ControlObject::doAdjustCursors()
{
	int i;

	Analog->Cursor = Analog->Cursor.scanAfter(cursorStamp());
	if (!Analog->Cursor.valid())
		--Analog->Cursor;

	for (i = 0; i <= maxBinUsed; i++) {
		BinaryControl *b = Binary + i;
		if (!b->Used)
			continue;

		b->Cursor = b->Cursor.scanAfter(cursorStamp());
		if (!b->Cursor.valid())
			--b->Cursor;
	}
}

unsigned long long ControlObject::beginUndefinedRegion(unsigned long long ts)
{
	unsigned int i;
	ARecord *a;

	if (!Online.end_ts)
		return 0;

	if (!ts)
		ts = Online.end_ts + 1;
	qWarning("undefined region started at %llu", ts);
	a = new AUndefinedRecord();
	newARecord(ts, a);
	ACC->invalidate_all();

	for (i = 0; i < nBinaries; i++)
		if (!Online.get_binary(i)->AlwaysDefined &&
				Online.get_binary(i)->Used)
			newBRecord(ts, i, Undefined);
	return ts;
}

void ControlObject::ping()
{
	DataTimer->start(AR3CMON_DATA_TIMEOUT, TRUE);
	if (HaveData)
		return;
	HaveData = true;
	emit sigHaveDataChanged();
}

void ControlObject::noData()
{
	HaveData = false;
	emit sigHaveDataChanged();
}

void ControlObject::clearOfflineData()
{
	if (OfflineMode)
		ACC->invalidate_all();
	Offline.clear();
}

void ControlObject::clearOnlineData()
{
	Online.clear();
	TimeStamp = 0;

	if (!ReviewMode) {
		setCursorStamp(0);
		noData();
	}
	if (!OfflineMode)
		ACC->invalidate_all();
}

void ControlObject::clearData()
{
	Online.clear();
	Offline.clear();
	TimeStamp = 0;
	setCursorStamp(0);

	enterNormalMode();
	noData();
}

void ControlObject::setAlwaysDefined(int id)
{
	binary(id)->AlwaysDefined = true;
}

void ControlObject::clearAlwaysDefined(int id)
{
	binary(id)->AlwaysDefined = false;
}

void ControlObject::__garbageCollector(void)
{
	int can_online = 1, can_offline = 1, can_offline_force = 1;
	int online_p, offline_p;

	if (mem_used_all() < mem_low)
		return;

	/* Ustawienie priorytetow dla danych offline i online
	 * priorytet > 1, wiksza liczba oznacza wikszy priotytet */
	if (OfflineMode) {
		online_p = 1;
		offline_p = 3;
	} else {
		online_p = 3;
		offline_p = 1;
	}

	while (mem_used_all() > mem_low) {
		if (can_online && can_offline) {
			if (Offline.get_used_all() / offline_p 
					> Online.get_used() / online_p) {
				if (!Offline.shrink(0))
					can_offline = 0;
			} else {
				if (!Online.shrink(0))
					can_online = 0;
			}
		} else if (can_online) {
			if (!Online.shrink(0))
				can_online = 0;
		} else if (can_offline) {
			if (!Offline.shrink(0))
				can_offline = 0;
		} else if (can_offline_force) {
			if (!Offline.shrink(1))
				can_offline_force = 0;
		} else if (!Offline.shrink_buffers(1))
			break;
	}
}

void ControlObject::garbageCollector()
{
	if (mem_used_all() > mem_high)
		__garbageCollector();
}

unsigned long sentinels = 0, items = 0, iters = 0, stores = 0;

void ControlObject::watchdog()
{
	Offline.watchdog();
}

void ControlObject::garbageCollector_stats()
{
	unsigned long used = Online.get_used();
	unsigned long off_used = Offline.get_used();
	unsigned long buff_used = Offline.get_used_bufs();

	fprintf(stderr, "Online: %lukB   Offline: %lukB   Buffers: %lukB   "
			"Total: %lukB   Low: %lukB   High: %lukB\n"
#ifdef OFFLINE_STATS
			"offline: reqs: %Lu, time = %.3fms, r/s: %.3f.\n"
#endif
			, used / 1024, off_used / 1024, buff_used / 1024,
			mem_used_all() / 1024, mem_low / 1024, mem_high / 1024
#ifdef OFFLINE_STATS
			, off_reqs, off_time / 1000.0, 1000000.0 * off_reqs
				/ (off_time ? off_time : 1)
#endif
			);
	ACC->stats();
}

void ControlObject::hashSetupData(unsigned char *buf, long len)
{
	unsigned long h = 0x65854ff3UL;

	if (len < 0) {
		SetupHash = 0;
		clearData();
		return;
	}

	for (long i = 0; i < len; i++)
		h = ((h + buf[i]) * 0x7f653f73UL + 0x35233745UL) % 0x7fffffffUL;
	qWarning("0x%08lx %ld", h, len);
	if (h == SetupHash) {
		qWarning("setup hash is OK, preserving data");
	} else {
		qWarning("setup hash is wrong, clearing data");
		SetupHash = h;
		clearData();
	}
}

void ControlObject::setCursor(unsigned long long ts)
{
	int move_right = ts > cursorStamp();

	ACC->invalidate_all();

	setCursorStamp(ts);
	enable_prefetch();
	if (move_right > 0) {
		Analog->Cursor = Analog->Cursor.scanRightAfter(cursorStamp());
		if (!Analog->Cursor.valid())
			--(Analog->Cursor);

		qWarning("scan righ: %lld/%lld", Analog->Cursor.timeStamp(),
				cursorStamp());

		for (int i = 0; i <= maxBinUsed; i++) {
			BinaryControl *b = Binary + i;
			if (!b->Used)
				continue;
			b->Cursor = b->Cursor.scanRightAfter(cursorStamp());
			if (!b->Cursor.valid())
				--(b->Cursor);
		}
	} else {
		Analog->Cursor = Analog->Cursor.scanLeftAfter(cursorStamp());
		if (!Analog->Cursor.valid())
			--(Analog->Cursor);

		qWarning("scan left: %lld/%lld", Analog->Cursor.timeStamp(),
				cursorStamp());
		for (int i = 0; i <= maxBinUsed; i++) {
			BinaryControl *b = Binary + i;
			if (!b->Used)
				continue;
			b->Cursor = b->Cursor.scanLeftAfter(cursorStamp());
			if (!b->Cursor.valid())
				--(b->Cursor);
		}
	}

	pleaseUpdate();
	emit sigUpdateTimer();
}

void ControlObject::advanceCursor(long long amount)
{
	setCursor(cursorStamp() + amount);
}

void ControlObject::setTimeRange(unsigned long long range)
{
	if (range < 1000000000ULL)
		range = 1000000000ULL;
	if (range > 3*60*60*1000000000ULL)
		range = 3*60*60*1000000000ULL;
	TimeRange = range;
	ACC->invalidate_all();
	autoAdjustVGrid();

	pleaseUpdate();
	emit sigUpdateTimer();
}

void ControlObject::autoAdjustVGrid()
{
	unsigned long long preferred = timeRange() / 8;
	static unsigned long long possible[] = {
		3*60*60*1000000000ULL, /*  3 hrs */
		2*60*60*1000000000ULL, /*  2 hrs */
		1*60*60*1000000000ULL, /*  1 hr */
		30*60*1000000000ULL, /* 30 min */
		20*60*1000000000ULL, /* 20 min */
		15*60*1000000000ULL, /* 15 min */
		10*60*1000000000ULL, /* 10 min */
		5*60*1000000000ULL, /*  5 min */
		3*60*1000000000ULL, /*  3 min */
		2*60*1000000000ULL, /*  2 min */
		1*60*1000000000ULL, /*  1 min */
		30*1000000000ULL, /* 30 sec */
		15*1000000000ULL, /* 15 sec */
		10*1000000000ULL, /* 10 sec */
		5*1000000000ULL, /*  5 sec */
		2*1000000000ULL, /*  2 sec */
		1*1000000000ULL, /*  1 sec */
		0ULL
	};
	unsigned long long *i = possible;

	if (preferred == 0)
		preferred = 1;
	do {
		++i;
	} while (*i >= preferred);

	i--;
	VGridStep = *i;
}

void ControlObject::end_of_data()
{
	Offline.finish_minute();
	nowCanDownload();
}

void ControlObject::nowCanDownload()
{
	pleaseUpdate();
}

void ControlObject::enterOfflineMode()
{
	active = &Offline;
	if (!OfflineMode)
		ACC->invalidate_all();
	Analog = Offline.get_acontrol();
	Binary = Offline.get_bcontrol();
	ReviewMode = true;
	OfflineMode = true;
	this->doAdjustCursors();
	emit sigModeChanged();
	UpdateTimer->start(ar3cmon_offline_update_interval);
	pleaseUpdate();
}

void ControlObject::enterNormalMode()
{
	active = &Online;
	if (ReviewMode)
		ACC->invalidate_all();
	Analog = Online.get_acontrol();
	Binary = Online.get_bcontrol();

	Online.cursors_eof();
	ReviewMode = false;
	OfflineMode = false;
	Online.set_cursor_stamp(Online.end_ts);
	this->doAdjustCursors();
	emit sigModeChanged();
	UpdateTimer->start(ar3cmon_update_interval);
	pleaseUpdate();
}

void ControlObject::enterOnlineMode()
{
	enterNormalMode();
	TheCommObject->goOnline();
}

void ControlObject::enterReviewMode()
{
	active = &Online;
	if (OfflineMode)
		ACC->invalidate_all();
	Analog = Online.get_acontrol();
	Binary = Online.get_bcontrol();
	ReviewMode = true;
	OfflineMode = false;
	this->doAdjustCursors();
	emit sigModeChanged();
	UpdateTimer->start(ar3cmon_offline_update_interval);
	pleaseUpdate();
}

void ControlObject::pleaseUpdate()
{
	emit sigPleaseUpdate();
	if (OfflineMode)
		Offline.please_update();
}

void ControlObject::reallyConnected()
{
	nowCanDownload();
}

void ControlObject::connected()
{
	Offline.connected();
}

void ControlObject::setBinaryUsed(int id)
{
	if (id < 0 || id >= 65535)
		return;

	onlineBinary(id)->Used = true;
	offlineBinary(id)->Used = true;
	if (id > maxBinUsed)
		maxBinUsed = id;
}

void ControlObject::newSyf(unsigned long long start, float current, float total)
{
	SyfStart = start;
	SyfCurrent = current;
	SyfTotal = total;
	emit sigSyfUpdate();
}

int ControlObject::penDriveConnected()
{
	//TODO
	//checking if pendrive is connected to USB port
	//if yes returns 1
	//else returns 0
	return 1;
}
	
void ControlObject::set_layer(int nlayer)
{
	if (layer != nlayer) {
		layer = nlayer;

		Offline.resync();
	}
}
	
void ControlObject::setValueFontSize(int size)
{
	ar3cmon_font_set_size(AR3CMONITOR_FONT_VALUE, size);
	ValueFont.setPointSize(size);
}

void ControlObject::setTimescaleFontSize(int size)
{
	ar3cmon_font_set_size(AR3CMONITOR_FONT_TIMESCALE, size);
	TimescaleFont.setPointSize(size);
}

void ControlObject::setFontSize(int size)
{
	ar3cmon_font_set_size(AR3CMONITOR_FONT, size);
}

void ControlObject::setSmallFontSize(int size)
{
	ar3cmon_font_set_size(AR3CMONITOR_FONT_SMALL, size);
	SmallFont.setPointSize(size);
}
	
void ControlObject::set_prec(prec_t p_prec)
{
	prec = p_prec;
}

prec_t ControlObject::get_prec(void)
{
	return prec;
}
	
void ControlObject::reset()
{
	prec = PREC_SDIGITS | 5;

	fonts_reset();
	ValueFont = make_font(AR3CMONITOR_FONT_VALUE);
	SmallFont = make_font(AR3CMONITOR_FONT_SMALL);
	TimescaleFont = make_font(AR3CMONITOR_FONT_TIMESCALE);
}
